/*
 * Example.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Example
{
    //
    // FIXME: Always change this GUID.
    //
    [ObjectId("6b4f29fc-c7e9-4dbc-b838-909ae6708f5d")]
    internal static class Program
    {
        #region Private Constants
        private static readonly string fieldName = "mainArgs";
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Data
        //
        // NOTE: Used for making random policy decisions.
        //
        private static Random random;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Public Data
        //
        // NOTE: Used to back the linked variable "mainArgs".
        //
        public static string[] mainArgs;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Application Entry Point
        private static void Main(string[] args)
        {
            //
            // NOTE: Save the command line arguments for use by the
            //       interpreter via the linked variable (optional).
            //
            mainArgs = args;

            //
            // NOTE: The "interpreter result" that is passed to various
            //       methods.
            //
            Result result = null;

            //
            // NOTE: First, we create a new interpreter (with the default
            //       options).
            //
            using (Interpreter interpreter = Interpreter.Create(args, ref result))
            {
                if (interpreter != null)
                {
                    ReturnCode code = interpreter.SetVariableLink(
                        VariableFlags.None, fieldName,
                        typeof(Program).GetField(fieldName), null, ref result);

                    if (code != ReturnCode.Ok)
                    {
                        //
                        // NOTE: Handle variable linking error.
                        //
                    }

                    //
                    // NOTE: Create one of our objects to handle our custom
                    //       command.
                    //
                    Class0 class0 = new Class0(new CommandData("class0", null,
                        null, ClientData.Empty, typeof(Class0).FullName,
                        CommandFlags.None, null, 0));

                    //
                    // NOTE: The token that will represent the custom command
                    //       we add.
                    //
                    long commandToken = 0;

                    //
                    // NOTE: Next, we can optionally add one or more custom
                    //       commands.
                    //
                    if (code == ReturnCode.Ok)
                        code = interpreter.AddCommand(class0, null,
                            ref commandToken, ref result);

                    //
                    // NOTE: The token that will represent the custom command
                    //       policy we add.
                    //
                    long policyToken = 0;

                    //
                    // NOTE: Next, add our custom command execution policy (for
                    //       use in "safe" mode).
                    //
                    if (code == ReturnCode.Ok)
                        code = interpreter.AddPolicy(Class0PolicyCallback,
                            null, null, ref policyToken, ref result);

                    //
                    // NOTE: The error line number that is passed to various
                    //       script evaluation methods.
                    //
                    int errorLine = 0;

                    //
                    // NOTE: Check for a successful return code.
                    //
                    if (code == ReturnCode.Ok) // OR: Utility.IsSuccess(code, true)
                    {
#if SHELL
                        result = null;

                        code = Interpreter.InteractiveLoop(interpreter, args,
                            ref result);
#else
                        //
                        // NOTE: Next, evaluate one or more scripts of your
                        //       choosing (which may or may not reference any
                        //       custom commands you may have added in the
                        //       previous step).
                        //
                        code = Engine.EvaluateScript(
                            interpreter, "class0 test; # <-- script text",
                            ref result, ref errorLine);
#endif

                        //
                        // NOTE: Check for an unsuccessful return code.
                        //
                        if (code != ReturnCode.Ok) // OR: !Utility.IsSuccess(code, true)
                        {
                            //
                            // NOTE: Handle script error.
                            //
                        }

                        //
                        // NOTE: Next, we can optionally remove one or more of
                        //       the custom command policies we added earlier.
                        //
                        code = interpreter.RemovePolicy(policyToken, null,
                            ref result);

                        //
                        // NOTE: Check for an unsuccessful return code.
                        //
                        if (code != ReturnCode.Ok)
                        {
                            //
                            // NOTE: Handle policy removal error.
                            //
                        }

                        //
                        // NOTE: Next, we can optionally remove one or more of
                        //       the custom commands we added earlier.
                        //
                        code = interpreter.RemoveCommand(commandToken, null,
                            ref result);

                        //
                        // NOTE: Check for an unsuccessful return code.
                        //
                        if (code != ReturnCode.Ok)
                        {
                            //
                            // NOTE: Handle command removal error.
                            //
                        }

                        code = interpreter.UnsetVariable(VariableFlags.None,
                            fieldName, ref result);

                        if (code != ReturnCode.Ok)
                        {
                            //
                            // NOTE: Handle variable unlinking error.
                            //
                        }
                    }
                    else
                    {
                        //
                        // NOTE: Handle failure to add the custom command
                        //       or policy.
                        //
                    }

                    //
                    // NOTE: Always check for a valid interpreter hosting
                    //       environment before using it as it is not
                    //       guaranteed to always be available.
                    //
                    IInteractiveHost interactiveHost = interpreter.Host;

                    if (interactiveHost != null)
                        interactiveHost.WriteResultLine(code, result, errorLine);
                    else
                        Console.WriteLine(Utility.FormatResult(
                            code, result, errorLine));
                }
                else
                {
                    //
                    // NOTE: Creation of the interpreter failed.
                    //
                    Console.WriteLine(Utility.FormatResult(
                        ReturnCode.Error, result));
                }
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Random Command Policy Implementation
        [MethodFlags(MethodFlags.CommandPolicy)]
        private static ReturnCode Class0PolicyCallback( /* POLICY */
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            //
            // NOTE: We only want to make policy decisions for our own
            //       command(s); therefore, use the appropriate helper
            //       method provided by the interpreter.
            //
            IPolicyContext policyContext = null;
            bool match = false;

            if (Utility.ExtractPolicyContextAndCommand(
                    interpreter, clientData, typeof(Class0), 0,
                    ref policyContext, ref match, ref result) == ReturnCode.Ok)
            {
                if (match)
                {
                    //
                    // NOTE: Create a new random policy decision maker to
                    //       allow us to easily demonstrate the effects of
                    //       both the "approved" and "denied" decisions.
                    //
                    if (random == null)
                        random = new Random();

                    //
                    // NOTE: Get a random integer; even numbers result in
                    //       a policy approval and odd numbers result in
                    //       policy denial.
                    //
                    if ((random.Next() % 2) == 0)
                        policyContext.Approved("random number is even");
                    else
                        policyContext.Denied("random number is odd");
                }

                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }
        #endregion
    }
}
