/*
 * Alias.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Commands
{
    [ObjectId("b338e2c4-6e66-456e-93af-91b5c21b449c")]
    /*
     * POLICY: This "command" is "safe" because it provides no
     *         functionality by itself (i.e. it is merely a
     *         transparent conduit to functionality that MAY
     *         be available elsewhere in the interpreter).
     */
    [CommandFlags(CommandFlags.Alias | CommandFlags.Safe)]
    [ObjectGroup("alias")]
    internal sealed class Alias : Core, IAlias
    {
        public Alias(
            ICommandData commandData
            )
            : base(commandData)
        {
            //
            // NOTE: This is not a strictly vanilla "command", it is a "command alias".
            //
            this.Kind |= IdentifierKind.Alias;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public Alias(
            ICommandData commandData,
            IAliasData aliasData
            )
            : this(commandData)
        {
            if (aliasData != null)
            {
                nameToken = aliasData.NameToken;
                targetInterpreter = aliasData.TargetInterpreter;
                target = aliasData.Target;
                arguments = aliasData.Arguments;
                options = aliasData.Options;
                startIndex = aliasData.StartIndex;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IAliasData Members
        private string nameToken;
        public string NameToken
        {
            get { return nameToken; }
            set { nameToken = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private Interpreter targetInterpreter;
        public Interpreter TargetInterpreter
        {
            get { return targetInterpreter; }
            set { targetInterpreter = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private IExecute target;
        public IExecute Target
        {
            get { return target; }
            set { target = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ArgumentList arguments;
        public ArgumentList Arguments
        {
            get { return arguments; }
            set { arguments = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private OptionDictionary options;
        public OptionDictionary Options
        {
            get { return options; }
            set { options = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int startIndex;
        public int StartIndex
        {
            get { return startIndex; }
            set { startIndex = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region System.Object Overrides
        public override string ToString()
        {
            return (arguments != null) ? arguments.ToString() : String.Empty;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IState Members
        public override ReturnCode Terminate(
            Interpreter interpreter,
            IClientData clientData,
            ref Result result
            )
        {
            ReturnCode code;

            if (interpreter != null)
            {
                if (nameToken != null)
                    //
                    // NOTE: The command is being deleted.  Delete the alias
                    //       as well.
                    //
                    code = interpreter.RemoveAlias(nameToken, clientData, ref result);
                else
                    //
                    // NOTE: The alias has already been deleted, skip it.
                    //
                    code = ReturnCode.Ok;

                if (code == ReturnCode.Ok)
                    code = base.Terminate(interpreter, clientData, ref result);
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IEnsemble Members
        public override EnsembleDictionary SubCommands
        {
            get
            {
                IEnsemble ensemble = target as IEnsemble;

                if (ensemble != null)
                {
                    return ensemble.SubCommands;
                }
                else
                {
                    if (targetInterpreter != null)
                    {
                        string targetName = null;

                        if (targetInterpreter.GetAliasArguments(this, arguments,
                                ref targetName) == ReturnCode.Ok)
                        {
                            //
                            // NOTE: Grab the target of the alias (this will be
                            //       null if we are late-bound).
                            //
                            IExecute aliasTarget = null;

                            if (targetInterpreter.GetAliasTarget(this,
                                    targetName, true, true, true,
                                    ref aliasTarget) == ReturnCode.Ok)
                            {
                                ensemble = aliasTarget as IEnsemble;

                                if (ensemble != null)
                                    return ensemble.SubCommands;
                            }
                        }
                    }
                }

                return null;
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IExecute Members
        public override ReturnCode Execute(
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code;

            if (targetInterpreter != null)
            {
                string targetName = null;
                ArgumentList targetArguments = null;

                code = targetInterpreter.GetAliasArguments(
                    this, arguments, ref targetName, ref targetArguments, ref result);

                if (code == ReturnCode.Ok)
                {
                    //
                    // NOTE: Grab the target of the alias (this will be
                    //       null if we are late-bound).
                    //
                    IExecute target = null;

                    code = targetInterpreter.GetAliasTarget(
                        this, targetName, true, true, true, ref target, ref result);

                    //
                    // NOTE: Do we have a valid target now (we may have
                    //       already had one or we may have just succeeded
                    //       in looking it up)?
                    //
                    if (code == ReturnCode.Ok)
                    {
                        //
                        // NOTE: Create and push a new call frame to track the
                        //       activation of this alias.
                        //
                        string name = StringList.MakeList("alias", this.Name);

                        ICallFrame frame = interpreter.NewLinkedCallFrame(
                            name, null, CallFrameFlags.Alias | CallFrameFlags.Tracking);

                        interpreter.PushCallFrame(frame);

                        if (FlagOps.HasFlags(this.Flags, CommandFlags.EvaluateAlias, true))
                        {
                            code = targetInterpreter.EvaluateScript(targetArguments, 0, ref result);
                        }
                        else
                        {
                            //
                            // NOTE: Save the current engine flags and then enable
                            //       the external execution flags.
                            //
                            EngineFlags savedEngineFlags = targetInterpreter.BeginExternalExecution();

                            code = targetInterpreter.Execute(
                                targetName, target, clientData, targetArguments, ref result);

                            //
                            // NOTE: Restore the saved engine flags, masking off the
                            //       external execution flags as necessary.
                            //
                            targetInterpreter.EndExternalExecution(savedEngineFlags);
                        }

                        //
                        // NOTE: Pop the original call frame that we pushed above
                        //       and any intervening scope call frames that may be
                        //       leftover (i.e. they were not explicitly closed).
                        //
                        interpreter.PopCallFrame(CallFrameFlags.Scope, true);
                    }
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion
    }
}

