/*
 * Callback.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Commands
{
    [ObjectId("7a66999b-d92f-4884-bf28-1ee7d7aa52ea")]
    [CommandFlags(CommandFlags.Safe)]
    [ObjectGroup("event")]
    internal sealed class Callback : Core
    {
        public Callback(
            ICommandData commandData
            )
            : base(commandData)
        {
            // do nothing.
        }

        #region IEnsemble Members
        private static readonly EnsembleDictionary subCommands = new EnsembleDictionary(new string[] {
            "clear", "dequeue", "enqueue", "execute"
        });

        public override EnsembleDictionary SubCommands
        {
            get { return subCommands; }
        }
        #endregion

        #region IExecute Members
        public override ReturnCode Execute(
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code;

            if (interpreter != null)
            {
                if (arguments != null)
                {
                    if (arguments.Count >= 2)
                    {
                        string subCommand = arguments[1];
                        bool tried = false;

                        code = RuntimeOps.TryExecuteSubCommandFromEnsemble(
                            this, interpreter, clientData, arguments, true,
                            false, ref subCommand, ref tried, ref result);

                        if ((code == ReturnCode.Ok) && !tried)
                        {
                            //
                            // NOTE: Programmatically interact with the debugger (breakpoint, watch,
                            //       eval, etc).
                            //
                            switch (subCommand)
                            {
                                case "clear":
                                    {
                                        if (arguments.Count == 2)
                                        {
                                            code = interpreter.ClearCallbackQueue(ref result);

                                            if (code == ReturnCode.Ok)
                                                result = String.Empty;
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"callback clear\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "dequeue":
                                    {
                                        if (arguments.Count >= 2)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-objectname", null),
                                                new Option(null, OptionFlags.MustHaveTypeValue, Index.Invalid, Index.Invalid, "-type", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-create", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nodispose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-alias", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-aliasall", null),
#if NATIVE && TCL
                                                new Option(null, OptionFlags.MustHaveTclInterpreterValue, Index.Invalid, Index.Invalid, "-tcl", null),
#endif
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-noforcedelete", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-tostring", null),
                                                new Option(typeof(ObjectFlags), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-objectflags",
                                                    new Variant(ObjectFlags.Default)),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            if (arguments.Count > 2)
                                                code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);
                                            else
                                                code = ReturnCode.Ok;

                                            if (code == ReturnCode.Ok)
                                            {
                                                if (argumentIndex == Index.Invalid)
                                                {
                                                    //
                                                    // NOTE: Do we expect to create an object (e.g. System.Int32.Parse)?
                                                    //
                                                    bool create = false;

                                                    if (options.IsPresent("-create"))
                                                        create = true;

                                                    bool dispose = true;

                                                    if (options.IsPresent("-nodispose"))
                                                        dispose = false;

                                                    //
                                                    // NOTE: Do we want to create a command alias for any newly created
                                                    //       objects?
                                                    //
                                                    bool alias = false;

                                                    if (options.IsPresent("-alias"))
                                                        alias = true;

                                                    bool aliasAll = false;

                                                    if (options.IsPresent("-aliasall"))
                                                        aliasAll = true;

                                                    //
                                                    // NOTE: Forcibly convert the object to a string and discard it?
                                                    //
                                                    bool toString = false;

                                                    if (options.IsPresent("-tostring"))
                                                        toString = true;

                                                    //
                                                    // NOTE: Which Tcl interpreter do we want a command alias created
                                                    //       in, if any?
                                                    //
                                                    Variant value = null;
                                                    Type type = null;

                                                    if (options.IsPresent("-type", ref value))
                                                        type = (Type)value.Value;

                                                    string interpName = null;

#if NATIVE && TCL
                                                    if (options.IsPresent("-tcl", ref value))
                                                        interpName = value.ToString();
#endif

                                                    ObjectFlags objectFlags = ObjectFlags.Default;

                                                    if (options.IsPresent("-objectflags", ref value))
                                                        objectFlags = (ObjectFlags)value.Value;

                                                    if (options.IsPresent("-noforcedelete"))
                                                        objectFlags &= ~ObjectFlags.ForceDelete;

                                                    string objectName = null;

                                                    if (options.IsPresent("-objectname", ref value))
                                                        objectName = value.ToString();

                                                    ICallback callback = null;

                                                    code = interpreter.DequeueCallback(ref callback, ref result);

                                                    if (code == ReturnCode.Ok)
                                                        code = MarshalOps.FixupReturnValue(
                                                            interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                            type, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                            objectName, interpName, callback, create, dispose, alias,
                                                            aliasAll, toString, ref result);
                                                }
                                                else
                                                {
                                                    result = "wrong # args: should be \"callback dequeue ?options?\"";
                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"callback dequeue ?options?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "enqueue":
                                    {
                                        if (arguments.Count >= 3)
                                        {
                                            ICallback callback = CommandCallback.Create(
                                                CallbackFlags.Default, interpreter, clientData,
                                                new StringList(arguments, 2), ref result);

                                            if (callback != null)
                                            {
                                                code = interpreter.EnqueueCallback(callback, ref result);

                                                if (code == ReturnCode.Ok)
                                                    result = String.Empty;
                                            }
                                            else
                                            {
                                                code = ReturnCode.Error;
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"callback enqueue name ?arg ...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "execute":
                                    {
                                        if (arguments.Count == 2)
                                        {
                                            code = interpreter.ExecuteCallbackQueue(ref result);
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"callback execute\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                default:
                                    {
                                        result = RuntimeOps.BadSubCommand(null, null, subCommand, this, null, null);
                                        code = ReturnCode.Error;
                                        break;
                                    }
                            }
                        }
                    }
                    else
                    {
                        result = "wrong # args: should be \"callback option ?arg ...?\"";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    result = "invalid argument list";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion
    }
}
