/*
 * HelpOps.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

#if SHELL
using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
#endif

using Eagle._Attributes;

#if SHELL
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Containers.Private;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;
#endif

#if !CONSOLE && SHELL
using ConsoleColor = Eagle._Components.Public.ConsoleColor;
#endif

namespace Eagle._Components.Private
{
    [ObjectId("35ce3166-9f74-460f-aa62-5cb27798ef66")]
    internal static class HelpOps
    {
        #region Private Constants
        #region Help Formatting
#if SHELL
        private const int OptionsPerLine = 3; /* magic */

#if INTERACTIVE_COMMANDS
        private const int CommandsPerLine = 6; /* magic */

        private const string commandOnlyFormat = "{0}{1}";
        private const string commandListFormat = "{0,X}{1}"; /* NOTE: 'X' is replaced by an integer. */
        private const string commandWithArgumentsFormat = "{0}{1} {2}";
        private const string commandAndDescriptionFormat = "{0}{1} -- {2}";
        private const string commandWithArgumentsAndDescriptionFormat = "{0}{1} {2} -- {3}";
#endif

        private const string bannerFormat = "{0} v{1} {2}{3} {4} ({5}){6}";
        private const string bannerTextFormat = " {0}";
        private const string bannerReleaseFormat = "{0}";

        private const string optionFormat = "{0,X}"; /* NOTE: 'X' is replaced by an integer. */
#endif
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Data
#if SHELL && INTERACTIVE_COMMANDS
        private static readonly object syncRoot = new object();
        private static StringPairDictionary commandHelp = null;
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Help Support Methods
#if SHELL && INTERACTIVE_COMMANDS
        private static string FormatHelpItem( /* CANNOT RETURN NULL */
            Interpreter interpreter,   /* in */
            StringPairDictionary help, /* in */
            string helpKey,            /* in */
            bool summary               /* in */
            )
        {
            string result = null;

            if (help != null)
            {
                if (helpKey != null)
                {
                    string prefix = RuntimeOps.InteractiveCommandPrefix;

                    if (!String.IsNullOrEmpty(prefix) && helpKey.StartsWith(
                            prefix, StringOps.SystemNoCaseStringComparisonType))
                        helpKey = helpKey.Substring(prefix.Length);

                    if (help.ContainsKey(helpKey))
                    {
                        IPair<string> helpItem = help[helpKey];

                        if (helpItem != null)
                        {
                            if ((helpItem.X != null) && (helpItem.Y != null))
                            {
                                if (summary)
                                    result = String.Format(
                                        commandWithArgumentsFormat,
                                        prefix, helpKey, helpItem.X);
                                else
                                    result = String.Format(
                                        commandWithArgumentsAndDescriptionFormat,
                                        prefix, helpKey, helpItem.X, helpItem.Y);
                            }
                            else if (helpItem.X != null)
                            {
                                if (summary)
                                    result = String.Format(
                                        commandWithArgumentsFormat,
                                        prefix, helpKey, helpItem.X);
                                else
                                    result = String.Format(
                                        commandAndDescriptionFormat,
                                        prefix, helpKey, helpItem.X);
                            }
                            else if (helpItem.Y != null)
                            {
                                if (summary)
                                    result = String.Format(
                                        commandOnlyFormat,
                                        prefix, helpKey);
                                else
                                    result = String.Format(
                                        commandAndDescriptionFormat,
                                        prefix, helpKey, helpItem.Y);
                            }
                            else
                            {
                                if (summary)
                                    result = String.Format(
                                        commandOnlyFormat,
                                        prefix, helpKey);
                                else
                                    result = String.Format(
                                        "No help is available for command \"{0}\".",
                                        helpKey);
                            }
                        }
                        else
                        {
                            result = String.Format(
                                "Invalid help item for command \"{0}\".",
                                helpKey);
                        }
                    }
                    else
                    {
                        IExecute execute = null;

                        if (interpreter.GetIExecuteViaResolvers(
                                interpreter.GetResolveEngineFlags(true), helpKey,
                                null, true, true, ref execute) == ReturnCode.Ok)
                        {
                            ISyntax syntax = execute as ISyntax;

                            if (syntax != null)
                                result = syntax.Syntax;
                        }

                        //
                        // NOTE: If no syntax help was available or the command
                        //       was not found we simply return a suitable error
                        //       message.
                        //
                        if (result == null)
                            result = String.Format(
                                "There is no help item or command \"{0}\".",
                                helpKey);
                    }
                }
                else
                {
                    result = "Invalid help key.";
                }
            }
            else
            {
                result = "No help available.";
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static StringList GetInteractiveCommandNames(
            string pattern,
            bool noCase
            )
        {
            StringPairDictionary help = GetInteractiveHelp();

            if (help != null)
            {
                StringList result = new StringList(); /* COPY */

                ReturnCode code;
                Result error = null;

                code = GenericOps<string>.FilterList(
                    new StringList(help.Keys), result, 0, Index.Invalid,
                    ToStringFlags.None, pattern, noCase, ref error);

                if (code == ReturnCode.Ok)
                    return result;
                else
                    DebugOps.Complain(code, error);
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static StringPair GetInteractiveCommandHelp(
            string name
            )
        {
            if (name != null)
            {
                StringPairDictionary help = GetInteractiveHelp();
                IPair<string> result;

                if ((help != null) &&
                    help.TryGetValue(name, out result) &&
                    (result != null))
                {
                    return new StringPair(result.X, result.Y); /* COPY */
                }
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static StringPairDictionary GetInteractiveHelp() /* CANNOT RETURN NULL */
        {
            StringPairDictionary result = null;

            lock (syncRoot)
            {
                if (commandHelp == null)
                {
                    string flagHelp = String.Format(
                        "Where \"flags\" is a list of zero or more values, separated by spaces or " +
                        "commas, which may be prefixed by \"{0}\" (add), \"{1}\" (remove), \"{2}\" " +
                        "(set), or \"{3}\" (set, then add).  If no prefix is specified, \"{4}\" " +
                        "(set, then add) is assumed.  Available flags are: {5}.", EnumOps.AddFlagOperator,
                        EnumOps.RemoveFlagOperator, EnumOps.SetFlagOperator, EnumOps.SetAddFlagOperator,
                        EnumOps.DefaultFlagOperator, "{0}");

                    string createFlagsHelpItem = String.Format(flagHelp,
                        GenericOps<string>.ListToEnglish(
                            new StringSortedList(Enum.GetNames(typeof(CreateFlags))),
                            String.Empty, ", ", "and "));

                    string scriptFlagsHelpItem = String.Format(flagHelp,
                        GenericOps<string>.ListToEnglish(
                            new StringSortedList(Enum.GetNames(typeof(ScriptFlags))),
                            String.Empty, ", ", "and "));

                    string engineFlagsHelpItem = String.Format(flagHelp,
                        GenericOps<string>.ListToEnglish(
                            new StringSortedList(Enum.GetNames(typeof(EngineFlags))),
                            String.Empty, ", ", "and "));

                    string interpreterFlagsHelpItem = String.Format(flagHelp,
                        GenericOps<string>.ListToEnglish(
                            new StringSortedList(Enum.GetNames(typeof(InterpreterFlags))),
                            String.Empty, ", ", "and "));

                    string substitutionFlagsHelpItem = String.Format(flagHelp,
                        GenericOps<string>.ListToEnglish(
                            new StringSortedList(Enum.GetNames(typeof(SubstitutionFlags))),
                            String.Empty, ", ", "and "));

                    string eventFlagsHelpItem = String.Format(flagHelp,
                        GenericOps<string>.ListToEnglish(
                            new StringSortedList(Enum.GetNames(typeof(EventFlags))),
                            String.Empty, ", ", "and "));

                    string expressionFlagsHelpItem = String.Format(flagHelp,
                        GenericOps<string>.ListToEnglish(
                            new StringSortedList(Enum.GetNames(typeof(ExpressionFlags))),
                            String.Empty, ", ", "and "));

                    string headerFlagsHelpItem = String.Format(flagHelp,
                        GenericOps<string>.ListToEnglish(
                            new StringSortedList(Enum.GetNames(typeof(HeaderFlags))),
                            String.Empty, ", ", "and "));

                    string pluginFlagsHelpItem = String.Format(flagHelp,
                        GenericOps<string>.ListToEnglish(
                            new StringSortedList(Enum.GetNames(typeof(PluginFlags))),
                            String.Empty, ", ", "and "));

#if NOTIFY || NOTIFY_OBJECT
                    string notifyTypesHelpItem = String.Format(flagHelp,
                        GenericOps<string>.ListToEnglish(
                            new StringSortedList(Enum.GetNames(typeof(NotifyType))),
                            String.Empty, ", ", "and "));

                    string notifyFlagsHelpItem = String.Format(flagHelp,
                        GenericOps<string>.ListToEnglish(
                            new StringSortedList(Enum.GetNames(typeof(NotifyFlags))),
                            String.Empty, ", ", "and "));
#endif

                    commandHelp = new StringPairDictionary();

                    commandHelp.Add("about",
                        new StringPair(null,
                            "Displays version and license information."));

#if DEBUGGER
                    commandHelp.Add("again",
                        new StringPair(null,
                            "Processes the previous interactive input again."));
#endif

                    commandHelp.Add("ainfo",
                        new StringPair(null,
                            "Displays the argument information for this interactive debuggging session."));

                    commandHelp.Add("args",
                        new StringPair(null,
                            "Displays the command line arguments as they were passed to this interactive debuggging session."));

#if DEBUGGER
                    commandHelp.Add("break",
                        new StringPair(null,
                            "Starts an interactive debugging session."));
#endif

#if CONSOLE
                    commandHelp.Add("cancel",
                        new StringPair(null,
                            "Initiates script cancellation for all interpreters."));
#endif

                    commandHelp.Add("canexit",
                        new StringPair(null,
                            "Toggle whether scripts are allowed to exit the application."));

                    commandHelp.Add("cflags",
                        new StringPair("?flags?",
                            "Displays or sets the creation flags for the interactive interpreter.  " + createFlagsHelpItem));

                    commandHelp.Add("chans",
                        new StringPair("?replace?", String.Format(
                            "Adds or replaces the standard channels \"{0}\", \"{1}\", and \"{2}\" for the interactive interpreter.",
                            StandardChannel.Input, StandardChannel.Output, StandardChannel.Error)));

                    commandHelp.Add("check",
                        new StringPair("?action? ?type?", String.Format(
                            "Checks the distribution site \"{0}\" to see if this is the latest build of the script engine.  " +
                            "Optionally, it can either fetch the latest release file of the specified type " +
                            "(e.g. \"source\", \"setup\", or \"binary\") or launch the external updater and then exit.",
                            GlobalState.GetAssemblyUri())));

                    commandHelp.Add("cinfo",
                        new StringPair(null,
                            "Displays the control information for the interactive interpreter, such as whether the " +
                            "script in progress is being canceled."));

#if CALLBACK_QUEUE
                    commandHelp.Add("clearq",
                        new StringPair(null,
                            "Clears the callback queue for the interactive interpreter."));
#endif

                    /* LOCAL RESULT */
                    commandHelp.Add("clearr",
                        new StringPair(null,
                            "Clears the local result.  The master result is untouched."));

                    commandHelp.Add("cmd",
                        new StringPair("?arg ...?",
                            "Executes the \"Command Processor\" (i.e. shell) configured for the operating system " +
                            "with the specified arguments."));

                    commandHelp.Add("color",
                        new StringPair(null,
                            "Toggle whether color is used on the console."));

                    commandHelp.Add("complaint",
                        new StringPair(null,
                            "Displays the previously stored complaint for the interactive interpreter."));

                    /* LOCAL/MASTER RESULT */
                    commandHelp.Add("copyr",
                        new StringPair(null,
                            "Copies the master result into the local result.  The master result is untouched."));

                    commandHelp.Add("cuinfo",
                        new StringPair(null,
                            "Displays the custom information for the host, if any."));

#if DEBUGGER
                    commandHelp.Add("dinfo",
                        new StringPair(null,
                            "Displays the debugger information for this interactive debuggging session."));
#endif

                    commandHelp.Add("dpath",
                        new StringPair(null,
                            "Displays the path information for the interactive interpreter."));

                    commandHelp.Add("eflags",
                        new StringPair("?flags?",
                            "Displays or sets the engine flags for the interactive interpreter.  " + engineFlagsHelpItem));

                    commandHelp.Add("einfo",
                        new StringPair(null,
                            "Displays the engine status information for the interactive interpreter."));

                    commandHelp.Add("eninfo",
                        new StringPair(null,
                            "Displays the entity summary information for the interactive interpreter."));

#if DEBUGGER
                    commandHelp.Add("eval",
                        new StringPair("arg ?arg ...?",
                            "Evaluates the specified arguments using the isolated debugger interpreter."));
#endif

                    commandHelp.Add("evflags",
                        new StringPair("?flags?",
                            "Displays or sets the event flags for the interactive interpreter.  " + eventFlagsHelpItem));

                    commandHelp.Add("exceptions",
                        new StringPair(null,
                            "Toggle how exceptional return codes are formatted on the console."));

                    commandHelp.Add("exflags",
                        new StringPair("?flags?",
                            "Displays or sets the expression flags for the interactive interpreter.  " + expressionFlagsHelpItem));

                    commandHelp.Add("exit",
                        new StringPair(null,
                            "Exits the interactive interpreter immediately."));

                    commandHelp.Add("finfo",
                        new StringPair(null,
                            "Displays the flags for the interactive interpreter and the flags that were passed to this " +
                            "interactive debuggging session."));

                    commandHelp.Add("fresc",
                        new StringPair(null,
                            "Toggle whether script cancellation flags are forcibly reset before interactive input."));

                    commandHelp.Add("fresh",
                        new StringPair(null,
                            "Toggle whether the halt flag is forcibly reset before interactive input."));

                    /* LOCAL/MASTER RESULT */
                    commandHelp.Add("fresr",
                        new StringPair(null,
                            "Resets the master and local results in-place."));

                    commandHelp.Add("frinfo",
                        new StringPair("level ?all?",
                            "Displays detailed information about the specified call frame."));

                    commandHelp.Add("go",
                        new StringPair(null,
                            "When actively debugging, continues execution of the script in progress."));

                    commandHelp.Add("halt",
                        new StringPair(null,
                            "When actively debugging, halts execution of the script in progress and breaks out of all nested " +
                            "interactive debugging sessions immediately."));

                    commandHelp.Add("hcancel",
                        new StringPair(null,
                            "Queues an asynchronous event that causes the host to cancel pending evaluations."));

                    commandHelp.Add("help",
                        new StringPair("?command?",
                            "Displays help for the specified interactive command or a list of available interactive commands."));

                    commandHelp.Add("hexit",
                        new StringPair(null,
                            "Queues an asynchronous event that causes the host to exit (which may or may not terminate the entire application)."));

                    commandHelp.Add("hflags",
                        new StringPair("?flags?",
                            "Displays or sets the header display flags.  " + headerFlagsHelpItem));

                    commandHelp.Add("hinfo",
                        new StringPair("?dimensions? ?formatting? ?colors? ?names?",
                            "Displays detailed information about the host for the interactive interpreter."));

#if HISTORY
                    commandHelp.Add("histfile",
                        new StringPair("?fileName?",
                            "Displays or sets the default file name to use when loading and saving the command history for the interactive interpreter."));

                    commandHelp.Add("histinfo",
                        new StringPair(null,
                            "Displays the command history for the interactive interpreter."));

                    commandHelp.Add("histclear",
                        new StringPair(null,
                            "Clears the command history for the interactive interpreter."));

                    commandHelp.Add("histload",
                        new StringPair("fileName",
                            "Loads the command history for the interactive interpreter."));

                    commandHelp.Add("histsave",
                        new StringPair("fileName",
                            "Saves the command history for the interactive interpreter."));
#endif

                    commandHelp.Add("ieflags",
                        new StringPair("?flags?",
                            "Displays or sets the interactive command engine flags for the interactive interpreter.  " + engineFlagsHelpItem));

                    commandHelp.Add("ievflags",
                        new StringPair("?flags?",
                            "Displays or sets the interactive command event flags for the interactive interpreter.  " + eventFlagsHelpItem));

                    commandHelp.Add("iexflags",
                        new StringPair("?flags?",
                            "Displays or sets the interactive command expression flags for the interactive interpreter.  " + expressionFlagsHelpItem));

                    commandHelp.Add("iflags",
                        new StringPair("?flags?",
                            "Displays or sets the instance flags for the interactive interpreter.  " + interpreterFlagsHelpItem));

                    commandHelp.Add("iinfo",
                        new StringPair(null,
                            "Displays information about the interactive interpreter."));

                    commandHelp.Add("isflags",
                        new StringPair("?flags?",
                            "Displays or sets the interactive command substitution flags for the interactive interpreter.  " + substitutionFlagsHelpItem));

                    commandHelp.Add("leflags",
                        new StringPair("?flags?",
                            "Displays or sets the local engine flags for the interactive interpreter.  " + engineFlagsHelpItem));

                    commandHelp.Add("levflags",
                        new StringPair("?flags?",
                            "Displays or sets the local event flags for the interactive interpreter.  " + eventFlagsHelpItem));

                    commandHelp.Add("lexflags",
                        new StringPair("?flags?",
                            "Displays or sets the local expression flags for the interactive interpreter.  " + expressionFlagsHelpItem));

                    commandHelp.Add("lfinfo",
                        new StringPair(null,
                            "Displays the local flags and the flags for the interactive interpreter."));

                    commandHelp.Add("lhflags",
                        new StringPair("?flags?",
                            "Displays or sets the local header display flags.  " + headerFlagsHelpItem));

                    /* LOCAL RESULT */
                    commandHelp.Add("lrinfo",
                        new StringPair(null,
                            "Displays detailed information about the local result, leaving it untouched.  " +
                            "The master result is untouched."));

                    commandHelp.Add("lsflags",
                        new StringPair("?flags?",
                            "Displays or sets the local substitution flags for the interactive interpreter.  " + substitutionFlagsHelpItem));

                    /* LOCAL/MASTER RESULT */
                    commandHelp.Add("mover",
                        new StringPair(null,
                            "Moves the local result into the master result.  The local result is reset."));

                    /* MASTER RESULT */
                    commandHelp.Add("mrinfo",
                        new StringPair("?previous?",
                            "Displays detailed information about the master result, leaving it untouched.  " +
                            "The local result is untouched."));

#if PREVIOUS_RESULT
                    /* LOCAL/PREVIOUS RESULT */
                    commandHelp.Add("nextr",
                        new StringPair(null,
                            "Sets the previous result for the interactive interpreter to the local result.  The master result is untouched."));
#endif

#if NOTIFY || NOTIFY_OBJECT
                    commandHelp.Add("nflags",
                        new StringPair("?flags?",
                            "Displays or sets the plugin notification flags.  " + notifyFlagsHelpItem));
#endif

                    commandHelp.Add("nop",
                        new StringPair(null,
                            "Does nothing.  The local result is untouched."));

#if NOTIFY || NOTIFY_OBJECT
                    commandHelp.Add("ntypes",
                        new StringPair("?flags?",
                            "Displays or sets the plugin notification types.  " + notifyTypesHelpItem));
#endif

                    /* LOCAL RESULT */
                    commandHelp.Add("nullr",
                        new StringPair(null,
                            "Sets the local result to null.  The master result is untouched."));

                    commandHelp.Add("oinfo",
                        new StringPair("name",
                            "Displays detailed information about the specified object."));

                    /* LOCAL RESULT */
                    commandHelp.Add("overr",
                        new StringPair("?options?",
                            "Sets the local result to the specified value.  The master result is untouched."));

                    commandHelp.Add("pflags",
                        new StringPair("?flags?",
                            "Displays or sets the plugin loader flags.  " + pluginFlagsHelpItem));

#if PREVIOUS_RESULT
                    /* LOCAL/PREVIOUS RESULT */
                    commandHelp.Add("prevr",
                        new StringPair(null,
                            "Sets the local result to the previous result for the interactive interpreter.  The master result is untouched."));
#endif

                    commandHelp.Add("ptest",
                        new StringPair("?pattern? ?all?",
                            "Runs one or more plugin tests for the interactive interpreter."));

                    commandHelp.Add("purge",
                        new StringPair(null,
                            "Purge undefined variables in the current call frame for the interactive interpreter."));

                    commandHelp.Add("queue",
                        new StringPair(null,
                            "Enter queued input mode.  The command or script entered will be queued for asynchronous evaluation."));

                    commandHelp.Add("rehash",
                        new StringPair("?profile?",
                            "Reloads the user-specific host profile."));

                    commandHelp.Add("resc",
                        new StringPair(null,
                            "Resets the cancel and unwind flags for the interactive interpreter."));

#if DEBUGGER
                    commandHelp.Add("reset",
                        new StringPair(null,
                            "Reset the state of the debugger to its initial default, enabled and inactive."));
#endif

                    commandHelp.Add("resh",
                        new StringPair(null,
                            "Resets the halt flag for the interactive interpreter."));

                    /* LOCAL/MASTER RESULT */
                    commandHelp.Add("resr",
                        new StringPair(null,
                            "Resets the master and local results."));

                    commandHelp.Add("restc",
                        new StringPair("?strict?",
                            "Restores the core plugin for the interactive interpreter."));

                    commandHelp.Add("restv",
                        new StringPair(null,
                            "Restores the core variables for the interactive interpreter."));

#if DEBUGGER
                    commandHelp.Add("resume",
                        new StringPair(null,
                            "When actively debugging, restores the debugger to its previous state."));
#endif

                    /* LOCAL/MASTER RESULT */
                    commandHelp.Add("rinfo",
                        new StringPair(null,
                            "Displays the master and local results, leaving them untouched."));

                    commandHelp.Add("rlimit",
                        new StringPair("?limit?",
                            "Displays or sets the recursion limit for the interactive interpreter."));

#if DEBUGGER
                    commandHelp.Add("run",
                        new StringPair(null,
                            "When actively debugging, disables single-stepping and continues execution of the script in progress."));
#endif

                    commandHelp.Add("scflags",
                        new StringPair("?flags?",
                            "Displays or sets the script flags for the interactive interpreter.  " + scriptFlagsHelpItem));

                    /* LOCAL/MASTER RESULT */
                    commandHelp.Add("setr",
                        new StringPair(null,
                            "Copies the local result into the master result.  The local result is untouched."));

                    commandHelp.Add("sflags",
                        new StringPair("?flags?",
                            "Displays or sets the substitution flags for the interactive interpreter.  " + substitutionFlagsHelpItem));

                    /* LOCAL/MASTER RESULT */
                    commandHelp.Add("show",
                        new StringPair("?local? ?empty?",
                            "Displays the information that corresponds to the local header display flags."));

                    commandHelp.Add("sinfo",
                        new StringPair(null,
                            "Displays the native stack information for the interactive interpreter."));

                    commandHelp.Add("stack",
                        new StringPair("?limit? ?all? ?info?",
                            "Displays the call stack for the interactive interpreter."));

#if DEBUGGER
                    commandHelp.Add("step",
                        new StringPair(null,
                            "Toggle whether single-stepping through scripts is enabled."));
#endif

                    commandHelp.Add("style",
                        new StringPair("?style?",
                            "Displays or sets the host output style for the debugger."));

#if DEBUGGER
                    commandHelp.Add("subst",
                        new StringPair("?-nobackslashes? ?-nocommands? ?-novariables? string",
                            "Substitutes the specified arguments using the isolated debugger interpreter."));

                    commandHelp.Add("suspend",
                        new StringPair(null,
                            "When actively debugging, saves the debugger state and then temporarily suspends debugging."));
#endif

#if NATIVE && TCL
                    commandHelp.Add("tclinterp",
                        new StringPair("?interp?",
                            "Displays or sets the the selected native Tcl interpreter for this interactive debuggging session, if available."));

                    commandHelp.Add("tclsh",
                        new StringPair(null,
                            "Toggle whether all scripts entered interactively are evaluated using the selected native Tcl interpreter, if available."));
#endif

                    commandHelp.Add("tclshrc",
                        new StringPair("?arg ...?", String.Format(
                            "Executes the \"Text Editor\" configured for the operating system to edit the \"{0}\" script file with the specified arguments.",
                            TclVars.RunCommandsFileName)));

                    commandHelp.Add("test",
                        new StringPair("?pattern? ?all?",
                            "Runs one or more tests for the interactive interpreter."));

                    commandHelp.Add("testdir",
                        new StringPair("directory",
                            "Displays or sets the directory used when searching for test files matching a specific pattern."));

                    commandHelp.Add("testgc",
                        new StringPair("start", String.Format(
                             "Starts or stops the thread that collects garbage every {0} milliseconds.",
                             TestOps.gcSleepTime)));

                    commandHelp.Add("testinfo",
                        new StringPair(null,
                            "Displays the test suite information for the interactive interpreter."));

                    commandHelp.Add("tinfo",
                        new StringPair(null,
                            "Displays the variable trace information for this interactive debuggging session."));

                    commandHelp.Add("toinfo",
                        new StringPair(null,
                            "Displays the token information for this interactive debuggging session."));

                    commandHelp.Add("trustclr",
                        new StringPair(null,
                            "Clears the list of trusted directories for the interactive interpreter."));

                    commandHelp.Add("trustdir",
                        new StringPair("directory",
                            "Displays or adds to the list of trusted directories for the interactive interpreter."));

                    commandHelp.Add("usage",
                        new StringPair(null,
                            "Displays the command line syntax for the shell."));

                    commandHelp.Add("version",
                        new StringPair(null,
                            "Displays the version information for the shell."));

                    commandHelp.Add("vinfo",
                        new StringPair("name ?elements? ?links?",
                            "Displays detailed information about the specified variable."));

                    commandHelp.Add("vout",
                        new StringPair("?channel? ?enabled?",
                            "Enables, disables, or displays queued virtual output for a channel."));
                }

                result = new StringPairDictionary(commandHelp);
            }

            return result;
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if SHELL
        private static bool HostTryWriteColor(
            IHost host,
            string value,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            bool wrote = false;

            if (host != null)
            {
                try
                {
                    if (FlagOps.HasFlags(host.GetFlags(), HostFlags.AllColors, false))
                    {
                        if (newLine)
                            wrote = host.WriteLine(
                                value, foregroundColor, backgroundColor);
                        else
                            wrote = host.Write(
                                value, foregroundColor, backgroundColor);
                    }
                }
                catch
                {
                    // do nothing.
                }

                if (!wrote)
                    wrote = newLine ? host.WriteLine(value) : host.Write(value);
            }

            return wrote;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Public Help Support Methods
#if SHELL
        public static bool WriteBanner(
            Interpreter interpreter,
            bool noRelease,
            bool noDescription
            )
        {
            if (interpreter != null)
            {
                IHost host = interpreter.Host;

                if (host != null)
                {
                    ConsoleColor foregroundColor = _ConsoleColor.Default;
                    ConsoleColor backgroundColor = _ConsoleColor.Default;

                    interpreter.GetHostColors(
                        host, ColorName.Banner, ref foregroundColor, ref backgroundColor);

                    Assembly assembly = GlobalState.GetAssembly();

                    string text = AttributeOps.GetAssemblyText(assembly);

                    string value = String.Format(bannerFormat,
                        GlobalState.GetPackageName(false),
                        GlobalState.GetAssemblyVersion(),
                        AttributeOps.GetAssemblyTag(assembly),
                        !String.IsNullOrEmpty(text) ?
                            String.Format(bannerTextFormat, text) : String.Empty,
                        FormatOps.PackageDateTime(
                            AttributeOps.GetAssemblyDateTime(assembly)),
                        AttributeOps.GetAssemblyConfiguration(assembly),
                        RuntimeOps.GetCertificateSubject(assembly, true));

                    if (!HostTryWriteColor(host, value, true, foregroundColor, backgroundColor))
                        return false;

                    if (!noRelease)
                    {
                        string release = AttributeOps.GetAssemblyRelease(assembly);

                        if (!String.IsNullOrEmpty(release))
                        {
                            value = String.Format(bannerReleaseFormat, release);

                            if (!host.WriteLine())
                                return false;

                            if (!HostTryWriteColor(host, value, true, foregroundColor, backgroundColor))
                                return false;
                        }
                    }

                    if (!noDescription)
                    {
                        value = String.Format(
                            Vars.PackageDescription, TclVars.VersionValue);

                        if (!host.WriteLine())
                            return false;

                        if (!HostTryWriteColor(host, value, true, foregroundColor, backgroundColor))
                            return false;
                    }

                    return true;
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool WriteLegalese(
            Interpreter interpreter,
            bool summaryOnly
            )
        {
            if (interpreter != null)
            {
                IHost host = interpreter.Host;

                if (host != null)
                {
                    ConsoleColor foregroundColor = _ConsoleColor.Default;
                    ConsoleColor backgroundColor = _ConsoleColor.Default;

                    interpreter.GetHostColors(
                        host, ColorName.Legal, ref foregroundColor, ref backgroundColor);

                    Assembly assembly = GlobalState.GetAssembly();

                    string copyright = AttributeOps.GetAssemblyCopyright(
                        assembly, true);

                    bool wrote = false;

                    if (!String.IsNullOrEmpty(copyright))
                    {
                        if (HostTryWriteColor(host,
                                copyright, true, foregroundColor, backgroundColor))
                        {
                            wrote = true;
                        }
                        else
                        {
                            return false;
                        }
                    }

                    string license = AttributeOps.GetAssemblyLicense(assembly,
                        summaryOnly);

                    if (!String.IsNullOrEmpty(license))
                    {
                        if (wrote && !host.WriteLine())
                            return false;

                        if (HostTryWriteColor(host,
                                license, true, foregroundColor, backgroundColor))
                        {
                            wrote = true;
                        }
                        else
                        {
                            return false;
                        }
                    }

                    return true;
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode WriteUsage(
            Interpreter interpreter,
            string error,
            bool showBanner,
            bool showLegalese,
            bool showOptions,
            bool showEnvironment,
            ref Result result
            )
        {
            try
            {
                if (interpreter != null)
                {
                    IInteractiveHost interactiveHost = interpreter.Host;

                    if (interactiveHost != null)
                    {
                        //
                        // NOTE: Did the caller specify an error message to
                        //       display?
                        //
                        bool showError = !String.IsNullOrEmpty(error);

                        //
                        // HACK: Honor the quiet mode for the interpreter.
                        //
                        if (showError && interpreter.ShouldBeQuiet())
                            showError = false;

                        //
                        // NOTE: If we are going to display the error message
                        //       do it now, before any other output.
                        //
                        if (showError)
                            interactiveHost.WriteResultLine(
                                ReturnCode.Error, error);

                        if (showBanner)
                        {
                            if (showError)
                                interactiveHost.WriteLine();

                            WriteBanner(interpreter, false, false);
                        }

                        if (showLegalese)
                        {
                            if (showError || showBanner)
                                interactiveHost.WriteLine();

                            WriteLegalese(interpreter, true);
                        }

                        if (showOptions)
                        {
                            if (showError || showBanner || showLegalese)
                                interactiveHost.WriteLine();

                            interactiveHost.WriteLine(String.Format(
                                "usage:\n\n" +
                                "{0}{1} [options]",
                                Characters.HorizontalTab, Path.GetFileName(PathOps.GetExecutableName())));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine("options:");
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} <fileName> : Reads the entire specified file, attempting to\n" +
                                "{0}                        interpret each line as a list of arguments to\n" +
                                "{0}                        be inserted in order, replacing both the\n" +
                                "{0}                        \"-{1}\" option and the file name and then\n" +
                                "{0}                        continues processing arguments.  The literal\n" +
                                "{0}                        string \"{2}\" or \"{3}\" may be used to specify\n" +
                                "{0}                        the standard input stream as the file to read.",
                                Characters.HorizontalTab, CommandLineOption.Arguments,
                                CommandLineArgument.StandardInput, StandardChannel.Input));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} : Enables debug mode for the interpreter and then continues\n" +
                                "{0}         processing arguments (i.e. various strategically placed\n" +
                                "{0}         diagnostic messages are produced to help troubleshoot the\n" +
                                "{0}         startup process).",
                                Characters.HorizontalTab, CommandLineOption.Debug));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} <encodingName> : Sets the encoding to use for script files\n" +
                                "{0}                           and then continues processing arguments.",
                                Characters.HorizontalTab, CommandLineOption.Encoding));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} [string ...] : Evaluates the specified string(s) and then\n" +
                                "{0}                         exits.",
                                Characters.HorizontalTab, CommandLineOption.Evaluate));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} <fileName> [argument ...] : Evaluates the specified file and then\n" +
                                "{0}                                  exits.",
                                Characters.HorizontalTab, CommandLineOption.File));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} : Displays version and syntax information and then exits.",
                                Characters.HorizontalTab, CommandLineOption.Help));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} : Immediately attempts to initialize the script library for\n" +
                                "{0}              the interpreter and then continues processing arguments.",
                                Characters.HorizontalTab, CommandLineOption.Initialize));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} : Enables interactive mode for the interpreter and then\n" +
                                "{0}               continues processing arguments.",
                                Characters.HorizontalTab, CommandLineOption.Interactive));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} : Enters the interactive loop instead of exiting the process.",
                                Characters.HorizontalTab, CommandLineOption.NoExit));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} : Disables tracing and then continues processing arguments.",
                                Characters.HorizontalTab, CommandLineOption.NoTrace));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} : Waits for a key to be pressed and then continues processing\n" +
                                "{0}         arguments (useful for attaching a managed debugger before any\n" +
                                "{0}         scripts [with the possible exception of the script library]\n" +
                                "{0}         have been evaluated).",
                                Characters.HorizontalTab, CommandLineOption.Pause));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} [pattern] [all] [argument ...] : Runs the specified plugin\n" +
                                "{0}                                             test(s) or the full plugin\n" +
                                "{0}                                             test suite(s) if no\n" +
                                "{0}                                             pattern is supplied and\n" +
                                "{0}                                             then exits.",
                                Characters.HorizontalTab, CommandLineOption.PluginTest));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} <script> : Evaluates the specified script (after the\n" +
                                "{0}                           script library has been initialized) and\n" +
                                "{0}                           then continues processing arguments.  If the\n" +
                                "{0}                           script library has not been initialized, an\n" +
                                "{0}                           error is generated (see the \"-{2}\"\n" +
                                "{0}                           option, above).",
                                Characters.HorizontalTab, CommandLineOption.PostInitialize,
                                CommandLineOption.Initialize));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} <script> : Evaluates the specified script (before the\n" +
                                "{0}                          script library has been initialized) and then\n" +
                                "{0}                          continues processing arguments.  If the\n" +
                                "{0}                          script library has been initialized, an error\n" +
                                "{0}                          is generated (see the \"-{2}\" option,\n" +
                                "{0}                          above).",
                                Characters.HorizontalTab, CommandLineOption.PreInitialize,
                                CommandLineOption.Initialize));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} <profile> : Loads the specified host profile and then\n" +
                                "{0}                     continues processing arguments.",
                                Characters.HorizontalTab, CommandLineOption.Profile));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} <settings> : Loads the specified interpreter settings,\n"+
                                "{0}                          recreates the interpreter based on the new\n" +
                                "{0}                          settings, and then continues processing\n" +
                                "{0}                          arguments.",
                                Characters.HorizontalTab, CommandLineOption.Reconfigure));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} <optionName> : Adds, removes, or resets the specified\n" +
                                "{0}                              runtime option(s) and then continues\n" +
                                "{0}                              processing arguments.  If the script\n" +
                                "{0}                              library has not been initialized, an\n" +
                                "{0}                              error is generated (see the \"-{2}\"\n" +
                                "{0}                              option, above).",
                                Characters.HorizontalTab, CommandLineOption.RuntimeOption,
                                CommandLineOption.Initialize));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} : Enables \"safe\" mode for the interpreter and then continues\n" +
                                "{0}        processing arguments (all \"unsafe\" commands will be hidden).",
                                Characters.HorizontalTab, CommandLineOption.Safe));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} : Enables \"standard\" mode for the interpreter and then\n" +
                                "{0}            continues processing arguments (all \"non-standard\" commands\n" +
                                "{0}            will be hidden).",
                                Characters.HorizontalTab, CommandLineOption.Standard));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} : Enables single-step mode for the script debugger and then\n" +
                                "{0}        continues processing arguments.",
                                Characters.HorizontalTab, CommandLineOption.Step));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} [pattern] [all] [argument ...] : Runs the specified test(s) or\n" +
                                "{0}                                       the full test suite(s) if no\n" +
                                "{0}                                       pattern is supplied and then\n" +
                                "{0}                                       exits.",
                                Characters.HorizontalTab, CommandLineOption.Test));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} <directory> : Sets the base directory for the test\n" +
                                "{0}                             suite.  This directory is used when\n" +
                                "{0}                             searching for test files matching a\n" +
                                "{0}                             specific pattern.",
                                Characters.HorizontalTab, CommandLineOption.TestDirectory));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} : Enables tracing to all active trace listeners (including the\n" +
                                "{0}         console, if applicable) and then continues processing\n" +
                                "{0}         arguments.",
                                Characters.HorizontalTab, CommandLineOption.Trace));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} : Enables trace listener output to potentially be written\n" +
                                "{0}               to the host for the interpreter and then continues\n" +
                                "{0}               processing arguments.  This option is intended for use\n" +
                                "{0}               with custom shells.  The default shell does not make use\n" +
                                "{0}               of the interpreter flag set by this option.",
                                Characters.HorizontalTab, CommandLineOption.TraceToHost));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} <path> : Sets the vendor path (i.e. the name of an\n" +
                                "{0}                     additional sub-directory within each directory\n" +
                                "{0}                     searched when attempting to find user-specific\n" +
                                "{0}                     and/or application-specific files) and then\n" +
                                "{0}                     continues processing arguments.",
                                Characters.HorizontalTab, CommandLineOption.VendorPath));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}-{1} : Displays detailed version information and then exits.",
                                Characters.HorizontalTab, CommandLineOption.Version));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}The option names above are all case-insensitive.",
                                Characters.HorizontalTab));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If any unrecognized argument is encountered it will be passed to the\n" +
                                "{0}shell argument callback, if any; otherwise, an error will be generated.",
                                Characters.HorizontalTab));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If no arguments are supplied, or if at any time there are no arguments\n" +
                                "{0}remaining to be processed, the interactive loop will be entered (unless\n" +
                                "{0}it has been disabled).",
                                Characters.HorizontalTab));
                        }

                        if (showEnvironment)
                        {
                            if (showError || showBanner || showLegalese || showOptions)
                                interactiveHost.WriteLine();

                            interactiveHost.WriteLine(String.Format(
                                "{0}Environment variable names may be case-sensitive (i.e. it depends on\n" +
                                "{0}the underlying operating system).",
                                Characters.HorizontalTab));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything],\n"+
                                "{0}internal calls into the garbage collector will always wait for all\n" +
                                "{0}pending finalizers to complete, even when run from a non-default\n" +
                                "{0}application domain context.  By default, waiting for all pending\n" +
                                "{0}finalizers to complete is disabled in non-default application domain\n" +
                                "{0}contexts in an attempt to prevent potential deadlocks.",
                                Characters.HorizontalTab, EnvVars.AlwaysWaitForGC));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], waits until a\n" +
                                "{0}key is pressed and then triggers a managed debugger break (useful for\n" +
                                "{0}attaching a managed debugger before any significant initialization has\n" +
                                "{0}been performed).",
                                Characters.HorizontalTab, EnvVars.Break));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], the console\n" +
                                "{0}window will be enabled.  This option is intended to be processed by\n" +
                                "{0}custom shells.  The default shell does not process this option (i.e. it\n" +
                                "{0}always attempts to use and/or create a console window).",
                                Characters.HorizontalTab, EnvVars.Console));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set, its value will be\n" +
                                "{0}used to alter or set the creation flags for the interpreter.  If the\n" +
                                "{0}value cannot be converted to creation flags, it will be ignored.",
                                Characters.HorizontalTab, EnvVars.CreateFlags));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], debug mode is\n" +
                                "{0}enabled immediately after creating the interpreter (i.e. various\n" +
                                "{0}strategically placed diagnostic messages are produced to help\n" +
                                "{0}troubleshoot the startup process).",
                                Characters.HorizontalTab, EnvVars.Debug));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], quiet\n" +
                                "{0}mode is enabled immediately after creating the interpreter (see the\n" +
                                "{0}\"{2}\" environment variable, below).",
                                Characters.HorizontalTab, EnvVars.DefaultQuiet, EnvVars.Quiet));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything],\n" +
                                "{0}interactive mode is enabled immediately after creating the interpreter.",
                                Characters.HorizontalTab, EnvVars.Interactive));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set, its value will\n" +
                                "{0}be used to alter or set the instance flags for the interpreter.  If the\n" +
                                "{0}value cannot be converted to instance flags, it will be ignored.",
                                Characters.HorizontalTab, EnvVars.InterpreterFlags));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], internal\n" +
                                "{0}calls into the garbage collector will be disabled.",
                                Characters.HorizontalTab, EnvVars.NeverGC));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything],\n" +
                                "{0}internal calls into the garbage collector will never wait for all\n" +
                                "{0}pending finalizers to complete.  By default, waiting for all pending\n" +
                                "{0}finalizers to complete is disabled in non-default application domain\n" +
                                "{0}contexts in an attempt to prevent potential deadlocks.",
                                Characters.HorizontalTab, EnvVars.NeverWaitForGC));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], the host\n" +
                                "{0}script cancellation interface will not be enabled (e.g. the Control-C\n" +
                                "{0}key being pressed will not trigger script cancellation).",
                                Characters.HorizontalTab, EnvVars.NoCancel));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], the console\n" +
                                "{0}output will not be in color.",
                                Characters.HorizontalTab, EnvVars.NoColor));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], the\n" +
                                "{0}console window will be disabled.  This option is intended to be\n" +
                                "{0}processed by custom shells.  The default shell does not process this\n" +
                                "{0}option (i.e. it always attempts to use and/or create a console window).",
                                Characters.HorizontalTab, EnvVars.NoConsole));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], the console\n" +
                                "{0}icon will not be changed.",
                                Characters.HorizontalTab, EnvVars.NoIcon));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], script\n" +
                                "{0}library initialization is skipped after creating the interpreter.",
                                Characters.HorizontalTab, EnvVars.NoInitialize));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], the\n" +
                                "{0}interactive loop will not be entered.",
                                Characters.HorizontalTab, EnvVars.NoLoop));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], the host\n" +
                                "{0}profile will not be loaded.",
                                Characters.HorizontalTab, EnvVars.NoProfile));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything],\n" +
                                "{0}exceptions will not be thrown when a disposed object is accessed.",
                                Characters.HorizontalTab, EnvVars.NoThrowOnDisposed));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], the console\n" +
                                "{0}title will not be changed.",
                                Characters.HorizontalTab, EnvVars.NoTitle));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], all tracing\n" +
                                "{0}will be disabled.",
                                Characters.HorizontalTab, EnvVars.NoTrace));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], assembly\n" +
                                "{0}files will not be checked for trust during the interpreter creation\n" +
                                "{0}process.",
                                Characters.HorizontalTab, EnvVars.NoTrusted));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], assembly\n" +
                                "{0}strong name signatures will not be verified during the interpreter\n" +
                                "{0}creation process.",
                                Characters.HorizontalTab, EnvVars.NoVerified));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], the host\n" +
                                "{0}will attempt to load the specified profile.",
                                Characters.HorizontalTab, EnvVars.Profile));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], potentially\n" +
                                "{0}important diagnostic messages about non-fatal (yet otherwise\n" +
                                "{0}unreportable) errors will not cause a modal message box to be displayed\n" +
                                "{0}when the originating interpreter [host] is unknown or unavailable.  Use\n" +
                                "{0}of this environment variable is not recommended (i.e. normally, it\n" +
                                "{0}would be preferable to determine the root cause of the message and fix\n" +
                                "{0}that instead).",
                                Characters.HorizontalTab, EnvVars.Quiet));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], enable \"safe\"\n" +
                                "{0}mode for the interpreter (all \"unsafe\" commands will be hidden).",
                                Characters.HorizontalTab, EnvVars.Safe));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], enable\n" +
                                "{0}\"standard\" mode for the interpreter (all \"non-standard\" commands will\n" +
                                "{0}be hidden).",
                                Characters.HorizontalTab, EnvVars.Standard));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], single-step\n" +
                                "{0}mode for the script debugger is enabled immediately after creating the\n" +
                                "{0}interpreter.",
                                Characters.HorizontalTab, EnvVars.Step));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], unhandled\n" +
                                "{0}exceptions are rethrown after being reported.",
                                Characters.HorizontalTab, EnvVars.Throw));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], tracing to\n" +
                                "{0}all active trace listeners (including the console, if applicable) will\n" +
                                "{0}be enabled before any significant initialization has been performed.",
                                Characters.HorizontalTab, EnvVars.Trace));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], trace\n" +
                                "{0}listener output may be sent to the host for the interpreter.  This\n" +
                                "{0}option is intended for use with custom shells.  The default shell does\n" +
                                "{0}not make use of the interpreter flag set by this option.",
                                Characters.HorizontalTab, EnvVars.TraceToHost));
                            interactiveHost.WriteLine();
                            interactiveHost.WriteLine(String.Format(
                                "{0}If the \"{1}\" environment variable is set [to anything], it will\n" +
                                "{0}be interpreted as the name of an additional sub-directory within each\n" +
                                "{0}directory searched when attempting to find user-specific and/or\n" +
                                "{0}application-specific files.",
                                Characters.HorizontalTab, EnvVars.VendorPath));
                        }

                        result = String.Empty;
                        return ReturnCode.Ok;
                    }
                    else
                    {
                        result = "interpreter host not available";
                    }
                }
                else
                {
                    result = "invalid interpreter";
                }
            }
            catch (Exception e)
            {
                result = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode WriteVersion(
            Interpreter interpreter,
            bool showBanner,
            bool showLegalese,
            ref Result result
            )
        {
            try
            {
                if (interpreter != null)
                {
                    IInteractiveHost interactiveHost = interpreter.Host;

                    if (interactiveHost != null)
                    {
                        if (showBanner)
                            WriteBanner(interpreter, false, false);

                        if (showLegalese)
                        {
                            if (showBanner)
                                interactiveHost.WriteLine();

                            WriteLegalese(interpreter, true);
                        }

                        PluginWrapperDictionary plugins;

                        lock (interpreter.SyncRoot)
                        {
                            //
                            // NOTE: Snapshot semantics.
                            //
                            plugins = interpreter.plugins;

                            if (plugins != null)
                                plugins = new PluginWrapperDictionary(plugins);
                        }

                        bool wrote = false;

                        if (plugins != null)
                        {
                            foreach (KeyValuePair<string, _Wrappers.Plugin> pair in plugins)
                            {
                                IPlugin plugin = pair.Value;

                                if (plugin != null)
                                {
                                    Result aboutResult = null;

                                    if ((plugin.About(
                                            interpreter, ref aboutResult) == ReturnCode.Ok) &&
                                        !String.IsNullOrEmpty(aboutResult))
                                    {
                                        if (!wrote)
                                        {
                                            if (showBanner || showLegalese)
                                                interactiveHost.WriteLine();

                                            interactiveHost.WriteLine("Loaded plugins:");
                                            interactiveHost.WriteLine();

                                            wrote = true;
                                        }

                                        interactiveHost.WriteLine(aboutResult);
                                    }
                                }
                            }
                        }

                        StringList options = DefineConstants.OptionList;

                        if (options != null)
                        {
                            int maximumLength = 0;

                            foreach (string option in options)
                            {
                                if (!String.IsNullOrEmpty(option) &&
                                    (option.Length > maximumLength))
                                {
                                    maximumLength = option.Length;
                                }
                            }

                            if (maximumLength > 0)
                            {
                                if (showBanner || showLegalese || wrote)
                                    interactiveHost.WriteLine();

                                interactiveHost.WriteLine("Compiled with options:");
                                interactiveHost.WriteLine();

                                int count = 0;

                                foreach (string option in options)
                                {
                                    if (!String.IsNullOrEmpty(option))
                                    {
                                        //
                                        // HACK: Right align the option within this column.
                                        //
                                        interactiveHost.Write(String.Format(
                                            optionFormat.Replace("X",
                                            (maximumLength + 2).ToString()), option));

                                        if ((++count % OptionsPerLine) == 0)
                                        {
                                            interactiveHost.WriteLine();
                                            count = 0;
                                        }
                                        else
                                        {
                                            interactiveHost.Write(Characters.Space);
                                        }
                                    }
                                }

                                if (count > 0)
                                {
                                    interactiveHost.WriteLine();
                                    count = 0;
                                }
                            }
                        }

                        result = String.Empty;
                        return ReturnCode.Ok;
                    }
                    else
                    {
                        result = "interpreter host not available";
                    }
                }
                else
                {
                    result = "invalid interpreter";
                }
            }
            catch (Exception e)
            {
                result = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if INTERACTIVE_COMMANDS
        public static ReturnCode WriteInteractiveHelp(
            Interpreter interpreter,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code;

            if (interpreter != null)
            {
                IHost host = interpreter.Host;

                if (host != null)
                {
                    StringPairDictionary help = GetInteractiveHelp();

                    if (help != null)
                    {
                        ConsoleColor helpForegroundColor = _ConsoleColor.Default;
                        ConsoleColor helpBackgroundColor = _ConsoleColor.Default;

                        interpreter.GetHostColors(host, ColorName.Help,
                            ref helpForegroundColor, ref helpBackgroundColor);

                        ConsoleColor helpItemForegroundColor = _ConsoleColor.Default;
                        ConsoleColor helpItemBackgroundColor = _ConsoleColor.Default;

                        interpreter.GetHostColors(host, ColorName.HelpItem,
                            ref helpItemForegroundColor, ref helpItemBackgroundColor);

                        if ((arguments != null) && (arguments.Count == 2))
                        {
                            // host.WriteLine();

                            HostTryWriteColor(
                                host, FormatHelpItem(interpreter, help, arguments[1], false),
                                true, helpItemForegroundColor, helpItemBackgroundColor);

                            // host.WriteLine();
                        }
                        else
                        {
                            string prefix = RuntimeOps.InteractiveCommandPrefix;

                            host.WriteLine();

                            HostTryWriteColor(host,
                                "WARNING: Please note that all of the following commands work best when invoked\n" +
                                "interactively (i.e. they cannot be easily or consistently invoked via script).",
                                true, helpForegroundColor, helpBackgroundColor);

                            host.WriteLine();

                            HostTryWriteColor(host, String.Format(
                                "Please type \"{0}help ?command?\" for more information on a specific command.",
                                prefix), true, helpForegroundColor, helpBackgroundColor);

                            host.WriteLine();

                            StringSortedList keys = new StringSortedList(help.Keys);

                            HostTryWriteColor(host, String.Format(
                                "The {0} available commands are: ", keys.Count),
                                true, helpForegroundColor, helpBackgroundColor);

                            host.WriteLine();

                            int maximumLength = 0;

                            foreach (KeyValuePair<string, string> pair in keys)
                                if (pair.Key.Length > maximumLength)
                                    maximumLength = pair.Key.Length;

                            int count = 0;

                            foreach (KeyValuePair<string, string> pair in keys)
                            {
                                //host.Write(FormatHelpItem(help, pair.Key, true), false,
                                //    helpItemForegroundColor, helpItemBackgroundColor);

                                //
                                // HACK: Right align the command name within this column.
                                //
                                HostTryWriteColor(host, String.Format(
                                    commandListFormat.Replace("X",
                                        (maximumLength - pair.Key.Length + 2).ToString()),
                                    prefix, pair.Key), false, helpItemForegroundColor,
                                    helpItemBackgroundColor);

                                if ((++count % CommandsPerLine) == 0)
                                {
                                    host.WriteLine();
                                    count = 0;
                                }
                                else
                                {
                                    host.Write(Characters.Space);
                                }
                            }

                            if (count > 0)
                            {
                                host.WriteLine();
                                count = 0;
                            }

                            // host.WriteLine();
                        }

                        //
                        // NOTE: If we managed to display any help at all, return Ok.
                        //
                        result = String.Empty;
                        code = ReturnCode.Ok;
                    }
                    else
                    {
                        result = "help not available";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    result = "interpreter host not available";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
#endif
#endif
        #endregion
    }
}
