###############################################################################
#
# helper.tcl -- Eagle Package for Tcl (Garuda)
#
# Extensible Adaptable Generalized Logic Engine (Eagle)
# Package Index File
#
# Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
#
# See the file "license.terms" for information on usage and redistribution of
# this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# RCS: @(#) $Id: $
#
###############################################################################

if {![package vsatisfies [package provide Tcl] 8.4]} then {
  error "need Tcl 8.4 or higher"
}

if {[catch {package present Eagle}] == 0} then {
  error "need native Tcl"
}

###############################################################################

namespace eval ::Garuda {
  #############################################################################
  #***************************** SHARED VARIABLES *****************************
  #############################################################################

  #
  # NOTE: Display diagnostic messages while starting up this package?  This is
  #       used by the code in the CLR assembly manager contained in this
  #       package.  This is also used by the package test suite.
  #
  variable verbose; # DEFAULT: false

  if {![info exists verbose]} then {
    set verbose false
  }

  #
  # NOTE: The Tcl command used to log warnings, errors, and other messages
  #       generated by the package.  This is used by the code in the CLR
  #       assembly manager contained in this package.  This is also used by the
  #       package test suite.
  #
  variable logCommand; # DEFAULT: [namespace current]::noLog

  if {![info exists logCommand]} then {
    set logCommand [namespace current]::noLog
  }

  #############################################################################
  #**************************** PACKAGE VARIABLES *****************************
  #############################################################################

  #
  # NOTE: Use the configured environment variables when searching for the Eagle
  #       CLR assembly?
  #
  variable useEnvironment; # DEFAULT: true

  if {![info exists useEnvironment]} then {
    set useEnvironment true
  }

  #
  # NOTE: The environment variable names to check when attempting to find the
  #       Eagle root directory.  This list is used during the assembly search
  #       process in the [setup] procedure.
  #
  variable envVars; # DEFAULT: "Eagle_Dll EAGLE EagleLkg"

  if {![info exists envVars]} then {
    set envVars [list Eagle_Dll EAGLE EagleLkg]
  }

  #
  # NOTE: Use the configured Windows registry keys when searching for the Eagle
  #       CLR assembly?
  #
  variable useRegistry; # DEFAULT: true

  if {![info exists useRegistry]} then {
    set useRegistry true
  }

  #
  # NOTE: The registry key where all the versions of Eagle installed on this
  #       machine (via the setup) can be found.
  #
  variable rootRegistryKeyName; # DEFAULT: HKEY_LOCAL_MACHINE\\Software\\Eagle

  if {![info exists rootRegistryKeyName]} then {
    set rootRegistryKeyName HKEY_LOCAL_MACHINE\\Software\\Eagle
  }

  #
  # NOTE: The Eagle build configurations we know about and support.  This list
  #       is used during the CLR assembly search process in the [setup]
  #       procedure (below).
  #
  variable assemblyConfigurations; # DEFAULT: {Debug Release ""}

  if {![info exists assemblyConfigurations]} then {
    set assemblyConfigurations [list]

    #
    # HACK: When running under the auspices of the Eagle test suite, make sure
    #       we try to select the matching build configuration [and suffix].
    #
    set assemblyConfiguration ""

    if {[info exists ::test_flags(-configuration)]} then {
      append assemblyConfiguration $::test_flags(-configuration)

      if {[info exists ::test_flags(-suffix)]} then {
        append assemblyConfiguration  $::test_flags(-suffix)
      }
    }

    if {[string length $assemblyConfiguration] > 0} then {
      lappend assemblyConfigurations $assemblyConfiguration
    }

    #
    # NOTE: Always add the default build configurations last.
    #
    lappend assemblyConfigurations Debug Release ""
  }

  #
  # NOTE: The possible file names for the Eagle CLR assembly.
  #
  variable assemblyFileNames; # DEFAULT: Eagle.dll

  if {![info exists assemblyFileNames]} then {
    set assemblyFileNames [list Eagle.dll]
  }

  #
  # NOTE: The name of the package we will provide to Tcl.
  #
  variable packageName; # DEFAULT: Garuda

  if {![info exists packageName]} then {
    set packageName [lindex [split [string trim [namespace current] :] :] 0]
  }

  #
  # NOTE: The name of the dynamic link library containing the native code for
  #       this package.
  #
  variable packageBinaryFileName; # DEFAULT: Garuda.dll

  if {![info exists packageBinaryFileName]} then {
    set packageBinaryFileName $packageName[info sharedlibextension]
  }

  #
  # NOTE: Start the CLR immediately upon loading the package?  This is used by
  #       the code in the CLR assembly manager contained in this package.
  #
  variable startClr; # DEFAULT: true

  if {![info exists startClr]} then {
    set startClr true
  }

  #
  # NOTE: Start the bridge between Eagle and Tcl immediately upon loading the
  #       package?  This is used by the code in the CLR assembly manager
  #       contained in this package.
  #
  variable startBridge; # DEFAULT: true

  if {![info exists startBridge]} then {
    set startBridge true
  }

  #
  # NOTE: The fully qualified path and file name for the Eagle CLR assembly to
  #       be loaded.  This is used by the code in the CLR assembly manager
  #       contained in this package.
  #
  variable assemblyPath; # DEFAULT: <unset>

  #
  # NOTE: The fully qualified type name of the CLR method(s) to execute within
  #       the Eagle CLR assembly.  This is used by the code in the CLR assembly
  #       manager contained in this package.
  #
  variable typeName; # DEFAULT: Eagle._Components.Public.NativePackage

  if {![info exists typeName]} then {
    set typeName Eagle._Components.Public.NativePackage
  }

  #
  # NOTE: The extra method flags to use when invoking the CLR methods.  Refer
  #       to the MethodFlags enumeration for full details.  This is used by the
  #       code in the CLR assembly manager contained in this package.
  #
  variable methodFlags; # DEFAULT: 0x0

  if {![info exists methodFlags]} then {
    set methodFlags 0x0
  }

  #
  # NOTE: The name of the CLR method to execute when starting up the bridge
  #       between Eagle and Tcl.  This is used by the code in the CLR assembly
  #       manager contained in this package.
  #
  variable startupMethodName; # DEFAULT: Startup

  if {![info exists startupMethodName]} then {
    set startupMethodName Startup
  }

  #
  # NOTE: The name of the CLR method to execute when issuing control directives
  #       to the bridge between Eagle and Tcl.  This is used by the code in the
  #       CLR assembly manager contained in this package.
  #
  variable controlMethodName; # DEFAULT: Control

  if {![info exists controlMethodName]} then {
    set controlMethodName Control
  }

  #
  # NOTE: The name of the managed method to execute when detaching a specific
  #       Tcl interpreter from the bridge between Eagle and Tcl.  This is used
  #       by the code in the CLR assembly manager contained in this package.
  #
  variable detachMethodName; # DEFAULT: Detach

  if {![info exists detachMethodName]} then {
    set detachMethodName Detach
  }

  #
  # NOTE: The name of the managed method to execute when completely shutting
  #       down the bridge between Eagle and Tcl.  This is used by the code in
  #       the CLR assembly manager contained in this package.
  #
  variable shutdownMethodName; # DEFAULT: Shutdown

  if {![info exists shutdownMethodName]} then {
    set shutdownMethodName Shutdown
  }

  #
  # NOTE: The user arguments to pass to all of the managed methods.  If this
  #       value is specified, it MUST be a well-formed Tcl list.  This is used
  #       by the code in the CLR assembly manager contained in this package.
  #
  variable methodArguments; # DEFAULT: NONE

  if {![info exists methodArguments]} then {
    set methodArguments [list]
  }

  #############################################################################
  #**************************** SHARED PROCEDURES *****************************
  #############################################################################

  proc noLog { string } {
    #
    # NOTE: Do nothing.  This will end up returning success to the native code
    #       that uses the configured log command.  Returning success from the
    #       configured log command means "yes, please log this to the attached
    #       debugger (and/or the system debugger) as well".  Returning an error
    #       from the configured log command will prevent this behavior.  Other
    #       than that, returning an error from the configured log command is
    #       completely harmless.
    #
  }

  #
  # NOTE: Also defined in and used by "all.tcl".
  #
  proc maybeFullName { command } {
    set which [namespace which $command]

    if {[string length $which] > 0} then {
      return $which
    }

    return $command
  }

  #
  # NOTE: Also defined in and used by "all.tcl".
  #
  proc isValidDirectory { path } {
    variable logCommand
    variable verbose

    if {$verbose} then {
      catch {
        set caller [maybeFullName [lindex [info level -1] 0]]

        eval $logCommand [list \
            "$caller: Checking for directory \"$path\"..."]
      }
    }

    #
    # NOTE: For now, just make sure the path refers to an existing directory.
    #
    return [expr {[string length $path] > 0 && [file exists $path] && \
        [file isdirectory $path]}]
  }

  #
  # NOTE: Also defined in and used by "all.tcl".
  #
  proc isValidFile { path } {
    variable logCommand
    variable verbose

    if {$verbose} then {
      catch {
        set caller [maybeFullName [lindex [info level -1] 0]]

        eval $logCommand [list \
            "$caller: Checking for file \"$path\"..."]
      }
    }

    #
    # NOTE: For now, just make sure the path refers to an existing file.
    #
    return [expr {[string length $path] > 0 && [file exists $path] && \
        [file isfile $path]}]
  }

  #############################################################################
  #**************************** UTILITY PROCEDURES ****************************
  #############################################################################

  proc getEnvironmentPathList { varNames } {
    global env

    set result [list]

    foreach varName $varNames {
      if {[info exists env($varName)]} then {
        set path [string trim $env($varName)]

        if {[isValidDirectory $path] || [isValidFile $path]} then {
          lappend result $path
        }
      }
    }

    return $result
  }

  proc getRegistryPathList { rootKeyName valueName } {
    set result [list]

    catch {
      package require registry; # NOTE: Tcl for Windows only.

      foreach keyName [registry keys $rootKeyName] {
        set subKeyName $rootKeyName\\$keyName

        if {[catch {string trim [registry get \
            $subKeyName $valueName]} path] == 0} then {
          if {[isValidDirectory $path] || [isValidFile $path]} then {
            lappend result $path
          }
        }
      }
    }

    return $result
  }

  proc probeBinaryFile { directory configuration fileName } {
    set path $directory; # maybe it is really a file?

    if {[isValidFile $path]} then {
      return $path
    }

    if {[string length $configuration] > 0} then {
      set path [file join $directory bin $configuration bin $fileName]

      if {[isValidFile $path]} then {
        return $path
      }

      set path [file join $directory bin $configuration $fileName]

      if {[isValidFile $path]} then {
        return $path
      }
    } else {
      set path [file join $directory bin $fileName]

      if {[isValidFile $path]} then {
        return $path
      }
    }

    return ""
  }

  proc findBinaryFile { directories configurations fileNames } {
    foreach directory $directories {
      foreach configuration $configurations {
        foreach fileName $fileNames {
          set path [probeBinaryFile $directory $configuration $fileName]

          if {[isValidFile $path]} then {
            return $path
          }
        }
      }
    }

    return ""
  }

  #############################################################################
  #************************ PACKAGE HELPER PROCEDURES *************************
  #############################################################################

  proc haveEagle { {varName ""} } {
    #
    # NOTE: Attempt to determine if Eagle has been loaded successfully and is
    #       currently available for use.  First, check that there is a global
    #       command named "eagle".  Second, make sure we can use that command
    #       to evaluate a trivial Eagle script that fetches the name of the
    #       script engine itself from the Eagle interpreter.  Finally, compare
    #       that result with "eagle" to make sure it is really Eagle.
    #
    if {[llength [info commands ::eagle]] > 0 && \
        [catch {::eagle {set ::tcl_platform(engine)}} result] == 0 && \
        [string equal -nocase eagle $result]} then {
      #
      # NOTE: Ok, it looks like Eagle is loaded and ready for use.  If the
      #       caller wants the version, use the specified variable name to
      #       store it in the context of the caller.
      #
      if {[string length $varName] > 0} then {
        upvar 1 $varName version
      }

      #
      # NOTE: Fetch the full version of Eagle script engine.
      #
      if {[catch {::eagle {set ::eagle_platform(patchLevel)}} \
          version] == 0} then {
        #
        # NOTE: Finally, verify that the result looks like a proper version
        #       number using a suitable regular expression.
        #
        if {[regexp -- {^\d+\.\d+\.\d+\.\d+$} $version]} then {
          return true
        }
      }
    }

    return false
  }

  #############################################################################
  #************************ PACKAGE STARTUP PROCEDURE *************************
  #############################################################################

  proc setupAndLoad { directory } {
    variable assemblyConfigurations
    variable assemblyFileNames
    variable assemblyPath
    variable envVars
    variable logCommand
    variable packageBinaryFileName
    variable packageName
    variable rootRegistryKeyName
    variable useEnvironment
    variable useRegistry
    variable verbose

    if {[info exists assemblyPath]} then {
      #
      # NOTE: The managed assembly path has been pre-configured by an external
      #       script; therefore, just use it verbatim.
      #
      if {$verbose} then {
        catch {
          set caller [maybeFullName [lindex [info level 0] 0]]

          eval $logCommand [list \
              "$caller: Using existing assembly path \"$assemblyPath\"..."]
        }
      }
    } else {
      #
      # NOTE: Build the list of directories to search for the managed assembly.
      #
      set directories [list]

      if {$useEnvironment} then {
        eval lappend directories [getEnvironmentPathList $envVars]
      }

      if {$useRegistry} then {
        eval lappend directories [getRegistryPathList $rootRegistryKeyName \
            Path]
      }

      if {$verbose} then {
        catch {
          set caller [maybeFullName [lindex [info level 0] 0]]

          eval $logCommand [list \
              "$caller: List of directories to search: $directories"]
        }
      }

      #
      # NOTE: Attempt to find the Eagle managed assembly file using the list of
      #       candidate directories.
      #
      set path [findBinaryFile $directories $assemblyConfigurations \
          $assemblyFileNames]

      if {[isValidFile $path]} then {
        set assemblyPath $path
      }

      #
      # NOTE: If no managed assembly path could be found, use the default one.
      #       This is very unlikely to result in the package being successfully
      #       loaded.
      #
      if {![info exists assemblyPath] || \
          [string length $assemblyPath] == 0} then {
        #
        # NOTE: Choose the first (default) managed assembly file name residing
        #       in the same directory as the package.
        #
        set assemblyPath [file join $directory [lindex $assemblyFileNames 0]]

        if {$verbose} then {
          catch {
            set caller [maybeFullName [lindex [info level 0] 0]]

            eval $logCommand [list \
                "$caller: Using default assembly path \"$assemblyPath\"..."]
          }
        }
      }
    }

    #
    # NOTE: Attempt to load the dynamic link library for the package now that
    #       the managed assembly path has been set [to something].
    #
    load [file join $directory $packageBinaryFileName] $packageName
  }

  #############################################################################
  #***************************** PACKAGE STARTUP ******************************
  #############################################################################

  #
  # NOTE: Arrange to have the "haveEagle" helper procedure exported from this
  #       namespace and imported into the global namespace.
  #
  set namespace [namespace current]; namespace export -clear haveEagle
  namespace eval :: [list namespace forget ::${namespace}::*]
  namespace eval :: [list namespace import -force ::${namespace}::haveEagle]

  #
  # NOTE: Save the package path for later use.
  #
  set packagePath [file dirname [info script]]

  #
  # NOTE: Attempt to setup and load the package right now.
  #
  setupAndLoad $packagePath
}
